/// <reference types="node" />
import { script as Script, address as Address, coin as Coin, tx as TX, keyring as KeyRing } from 'bcoin';
/**
 * Generate a redeem script, removing a name/key pair from the blockchain.
 * Validates `userPubkey` and `servicePubkey`.
 *
 * @param userPubkey The user's public key.
 * @param servicePubkey The service's public key.
 * @param alocktime An absolute lock time, in blocks.
 */
declare function genRedeemScript(userPubkey: Buffer, servicePubkey: Buffer, alocktime: number): Script;
/**
 * Generate a commit redeem script.
 * @param userPubkey The user's public key.
 * @param nonce A 256-bit buffer representing a nonce.
 * @param name A name of at most 64 characters composed of URL-safe characters.
 * @param locktime An absolute lock time, in blocks.
 */
declare function genCommitRedeemScript(userPubkey: Buffer, nonce: Buffer, name: string, locktime: number): Script;
/**
 * Generate a P2SH address from a redeem script.
 * @param redeemScript The script to use.
 */
declare function genP2shAddr(redeemScript: Script): Address;
/**
 * Encode `nonce`, `locktime`, and `name` into a buffer of length
 * `32 + 4 + 1 + name.length`.
 * @param nonce 256-bit nonce.
 * @param locktime Lock time of at most 500000000 blocks.
 * @param name Name of at most length 64 containing URL-safe characters.
 */
declare function serializeCommitData(nonce: Buffer, locktime: number, name: string): Buffer;
export interface ICommitData {
    nonce: Buffer;
    locktime: number;
    name: string;
}
/**
 * Convert bytestream `data` into human-readable form. The first 32 bytes are
 * the nonce, followed by 32 bytes representing the locktime, then a 4-byte
 * integer representing the name length, then up to 64 bytes representing the
 * name.
 * @param data Binary data to decode.
 */
declare function deserializeCommitData(data: Buffer): ICommitData;
/**
 * Generate a commit transaction.
 * @param coins Array of coins to fund the transaction.
 * @param name Name for name/key pair.
 * @param locktime Absolute locktime
 * @param commitFee Commit fee, in satoshis.
 * @param registerFee Registration fee, in satoshis.
 * @param escrowFee Escrow fee, in satoshis.
 * @param feeRate Fee rate, in satoshis/kilobyte.
 * @param userRing The user's key ring.
 * @param servicePubKey Service public key.
 */
declare function genCommitTx(coins: Coin[], name: string, locktime: number, commitFee: number, registerFee: number, escrowFee: number, feeRate: number, userRing: KeyRing, servicePubKey: Buffer): TX;
/**
 * Generate a lock transaction.
 * @param commitTX The corresponding commit transaction.
 * @param name The name to use.
 * @param upfrontFee The upfront fee in satoshis to use the service, as
 * determined by the service.
 * @param lockedFee Fee incentivizing registrar to provide service, as
 * determined by the service.
 * @param feeRate Fee rate in satoshi/KB.
 * @param userRing The user key ring.
 * @param servicePubKey Service public key.
 * @param locktime Absolute lock time in blocks.
 */
declare function genLockTx(commitTX: TX, name: string, upfrontFee: number, lockedFee: number, feeRate: number, userRing: KeyRing, servicePubKey: Buffer, locktime: number): TX;
/**
 * Attempt to extract a name from a lock transaction.
 * @param lockTx Lock transaction to read from.
 * @returns The name if present, and `null` otherwise.
 */
declare function getLockTxName(lockTx: TX): string | null;
/**
 * Attempt to extract a time lock from a lock transaction.
 * @param lockTx Lock transaction to read from.
 * @returns The lock time if present, and `null` otherwise.
 */
declare function getLockTxTime(lockTx: TX): number | null;
/**
 * Attempt to extract a public key from a lock transaction.
 * @param lockTx Lock transaction to read from.
 * @returns The public key if present, and `null` otherwise.
 */
declare function getLockTxPubKey(lockTx: TX): Buffer | null;
/**
 * Generate an unlock transaction.
 * @param lockTx The corresponding lock transaction.
 * @param commitTx The corresponding commit transaction.
 * @param feeRate The fee rate in satoshi/KB.
 * @param service Whether to use the script as service.
 * @param ring The service key ring if `service`, otherwise user key ring.
 * @param otherPubKey The user key ring if `service`, otherwise service key ring.
 */
declare function genUnlockTx(lockTx: TX, commitTx: TX, feeRate: number, service: boolean, ring: KeyRing, otherPubKey: Buffer): TX;
export { genLockTx, genUnlockTx, genRedeemScript, genP2shAddr, getLockTxName, getLockTxTime, getLockTxPubKey, genCommitTx, genCommitRedeemScript, serializeCommitData, deserializeCommitData };
