import { address, crypto, script, output, mtx } from 'bcoin';
import randomBytes from 'randombytes';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = Object.setPrototypeOf ||
    ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
    function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

/* tslint:disable:max-classes-per-file */
var BadUserPublicKeyError = /** @class */ (function (_super) {
    __extends(BadUserPublicKeyError, _super);
    function BadUserPublicKeyError() {
        var _newTarget = this.constructor;
        var _this = this;
        var trueProto = _newTarget.prototype;
        _this = _super.call(this, 'Invalid user public key') || this;
        // Alternatively use Object.setPrototypeOf if you have an ES6 environment.
        _this.__proto__ = trueProto;
        return _this;
    }
    return BadUserPublicKeyError;
}(Error));
var BadServicePublicKeyError = /** @class */ (function (_super) {
    __extends(BadServicePublicKeyError, _super);
    function BadServicePublicKeyError() {
        var _newTarget = this.constructor;
        var _this = this;
        var trueProto = _newTarget.prototype;
        _this = _super.call(this, 'Invalid service public key') || this;
        // Alternatively use Object.setPrototypeOf if you have an ES6 environment.
        _this.__proto__ = trueProto;
        return _this;
    }
    return BadServicePublicKeyError;
}(Error));
var BadLockTransactionError = /** @class */ (function (_super) {
    __extends(BadLockTransactionError, _super);
    function BadLockTransactionError() {
        var _newTarget = this.constructor;
        var _this = this;
        var trueProto = _newTarget.prototype;
        _this = _super.call(this, 'This locking transaction is not valid for this service') || this;
        // Alternatively use Object.setPrototypeOf if you have an ES6 environment.
        _this.__proto__ = trueProto;
        return _this;
    }
    return BadLockTransactionError;
}(Error));
/* tslint:enable:max-classes-per-file */

/**
 * Returns a Boolean value that indicates if URI has and only has alphanumeric leteral and  '_',  '-', '.' '~'
 * @param str The URI String
 */
function isURISafe(str) {
    var re = /^[a-zA-Z0-9_\-\.\~]*$/;
    return re.test(str);
}
/**
 * Returns a Boolean value indicates if transaction output is in the correct format of Opcode OP_RETURN.
 * @param output Transaction output
 */
function isValidOP_RETURN(output$$1) {
    // Check that output 0 is an OP_RETURN
    if (!output$$1.script.isNulldata()) {
        return false;
    }
    // Check that output 0 contains 0 satoshis
    if (output$$1.value !== 0) {
        return false;
    }
    // Check that output 0 contains exactly 2 opcodes
    if (output$$1.script.length !== 2) {
        return false;
    }
    return true;
}
/**
 * Returns a Boolean value indicates that whether or not the commit transaction is valid
 * @param tx The commit transaction that needs to be verified.
 * @param userPubKey User's public key
 * @param servicePubKey Service provider's public key
 * @param name A name of at most 64 characters composed of URL-safe characters
 * @param locktime An absolute lock time, in blockheight
 */
function verifyCommitTX(tx, userPubKey, servicePubKey, name, locktime) {
    if (tx.outputs.length < 3) {
        return false;
    }
    // Check that output 0 is an OP_RETURN of the correct form
    if (!isValidOP_RETURN(tx.outputs[0])) {
        return false;
    }
    // Check that output 0 contains a 32-byte nonce
    var nonce = tx.outputs[0].script.code[1].data;
    if (nonce.length !== 32) {
        return false;
    }
    // Check that output 1 is sent to the service's address
    var servicePKH = crypto.hash160(servicePubKey);
    var serviceAddr = address.fromPubkeyhash(servicePKH);
    if (tx.outputs[1].getAddress().toBase58('testnet') !== serviceAddr.toBase58('testnet')) {
        return false;
    }
    // Check that output 2 is a P2SH
    if (!tx.outputs[2].script.isScripthash()) {
        return false;
    }
    // Check that output 2 script is correct
    var redeemScript = genCommitRedeemScript(userPubKey, nonce, name, locktime);
    var scriptHash = crypto.hash160(redeemScript.toRaw());
    var p2shAddr = address.fromScripthash(scriptHash);
    if (tx.outputs[2].getAddress().toBase58('testnet') !== p2shAddr.toBase58('testnet')) {
        return false;
    }
    return true;
}
/**
 * Returns a Boolean value indicates that whether or not the lock transaction is valid
 * @param tx The lock transaction
 * @param commitTX The corresponding commit transaction.
 * @param servicePubKey service provider's public key
 */
function verifyLockTX(tx, commitTX, servicePubKey) {
    if (tx.outputs.length < 2) {
        return false;
    }
    // Check that input 0 contains a valid pubkey
    var pubKey = getLockTxPubKey(tx);
    if (pubKey === null || !crypto.secp256k1.publicKeyVerify(pubKey)) {
        return false;
    }
    // Check that input 0 name data is only 64 bytes in length
    var nameStr = getLockTxName(tx);
    if (nameStr === null) {
        return false;
    }
    if (nameStr.length > 64) {
        return false;
    }
    // Check that output 1 name data contains only URL-safe characters
    if (!isURISafe(nameStr)) {
        return false;
    }
    // Check that output 0 is a P2PKH
    if (!tx.outputs[0].script.isPubkeyhash()) {
        return false;
    }
    // Check that output 0 is sent to the service's address
    var servicePKH = crypto.hash160(servicePubKey);
    var serviceAddr = address.fromPubkeyhash(servicePKH);
    if (tx.outputs[0].getAddress().toBase58('testnet') !== serviceAddr.toBase58('testnet')) {
        return false;
    }
    // Check that output 1 is a P2SH
    if (!tx.outputs[1].script.isScripthash()) {
        return false;
    }
    // Check that output 1 script is correct
    var locktime = getLockTxTime(tx);
    var redeemScript = genRedeemScript(pubKey, servicePubKey, locktime);
    var scriptHash = crypto.hash160(redeemScript.toRaw());
    var p2shAddr = address.fromScripthash(scriptHash);
    if (tx.outputs[1].getAddress().toBase58('testnet') !== p2shAddr.toBase58('testnet')) {
        return false;
    }
    // Check that input 0 is a valid commit TX
    if (!verifyCommitTX(commitTX, pubKey, servicePubKey, nameStr, locktime)) {
        return false;
    }
    return true;
}

/**
 * Generate a redeem script, removing a name/key pair from the blockchain.
 * Validates `userPubkey` and `servicePubkey`.
 *
 * @param userPubkey The user's public key.
 * @param servicePubkey The service's public key.
 * @param alocktime An absolute lock time, in blocks.
 */
function genRedeemScript(userPubkey, servicePubkey, alocktime) {
    // Validate user public key
    if (!crypto.secp256k1.publicKeyVerify(userPubkey)) {
        throw new BadUserPublicKeyError();
    }
    // Validate service public key
    if (!crypto.secp256k1.publicKeyVerify(servicePubkey)) {
        throw new BadServicePublicKeyError();
    }
    var script$$1 = new script(null);
    script$$1.pushSym('OP_IF');
    //
    // If spending as user, execute this branch
    // Verify that 0 <= current block size - commit block size
    script$$1.pushInt(0);
    script$$1.pushSym('OP_CHECKSEQUENCEVERIFY');
    script$$1.pushSym('OP_DROP');
    // Check the provided user signature
    script$$1.pushData(userPubkey);
    script$$1.pushSym('OP_CHECKSIG');
    script$$1.pushSym('OP_ELSE');
    //
    // Otherwise, if spending as service, execute this branch
    // Verify that alocktime <= current block size
    script$$1.pushInt(alocktime);
    script$$1.pushSym('OP_CHECKLOCKTIMEVERIFY');
    script$$1.pushSym('OP_DROP');
    // Check the provided service signature
    script$$1.pushData(servicePubkey);
    script$$1.pushSym('OP_CHECKSIG');
    script$$1.pushSym('OP_ENDIF');
    script$$1.compile();
    return script$$1;
}
/**
 * Generate a commit redeem script.
 * @param userPubkey The user's public key.
 * @param nonce A 256-bit buffer representing a nonce.
 * @param name A name of at most 64 characters composed of URL-safe characters.
 * @param locktime An absolute lock time, in blocks.
 */
function genCommitRedeemScript(userPubkey, nonce, name, locktime) {
    // Validate user public key
    if (!crypto.secp256k1.publicKeyVerify(userPubkey)) {
        throw new BadUserPublicKeyError();
    }
    var script$$1 = new script();
    // Verify that at least six blocks have passed since commit
    script$$1.pushInt(6);
    script$$1.pushSym('OP_CHECKSEQUENCEVERIFY');
    script$$1.pushSym('OP_DROP');
    //
    // Hash [256-bit nonce + 2-byte locktime (BE) + 1-byte length of name + name]
    // and check against parameters.
    script$$1.pushSym('OP_HASH256');
    var hashData = serializeCommitData(nonce, locktime, name);
    var hash = crypto.hash256(hashData);
    script$$1.pushData(hash);
    script$$1.pushSym('OP_EQUALVERIFY');
    // Check user signature
    script$$1.pushData(userPubkey);
    script$$1.pushSym('OP_CHECKSIG');
    script$$1.compile();
    return script$$1;
}
/**
 * Generate a P2SH address from a redeem script.
 * @param redeemScript The script to use.
 */
function genP2shAddr(redeemScript) {
    var outputScript = script.fromScripthash(redeemScript.hash160());
    var p2shAddr = address.fromScript(outputScript);
    return p2shAddr;
}
/**
 * Encode `nonce`, `locktime`, and `name` into a buffer of length
 * `32 + 4 + 1 + name.length`.
 * @param nonce 256-bit nonce.
 * @param locktime Lock time of at most 500000000 blocks.
 * @param name Name of at most length 64 containing URL-safe characters.
 */
function serializeCommitData(nonce, locktime, name) {
    // Verify nonce size
    if (nonce.length !== 32) {
        throw new Error('Invalid nonce size');
    }
    // Verify lock time is within bounds
    if (locktime > 500000000) {
        throw new Error('Locktime must be less than 500000000 blocks');
    }
    // Verify name is within bounds
    if (name.length > 64) {
        throw new Error('Name is too long');
    }
    // Create a new buffer. Write no
    var outBuf = new Buffer(32 + 4 + 1 + name.length);
    nonce.copy(outBuf);
    outBuf.writeUInt32BE(locktime, 32);
    outBuf.writeUInt8(name.length, 36);
    outBuf.write(name, 37, name.length, 'ascii');
    return outBuf;
}
/**
 * Convert bytestream `data` into human-readable form. The first 32 bytes are
 * the nonce, followed by 32 bytes representing the locktime, then a 4-byte
 * integer representing the name length, then up to 64 bytes representing the
 * name.
 * @param data Binary data to decode.
 */
function deserializeCommitData(data) {
    if (data.length < 38) {
        throw new Error('Invalid commit data');
    }
    var nonce = data.slice(0, 32);
    var locktime = data.readUInt32BE(32);
    var nameLen = data.readUInt8(36);
    var nameRaw = data.slice(37);
    // Verify the specified name length equals the actual length
    if (nameRaw.length !== nameLen) {
        throw new Error('Name has incorrect length');
    }
    var name = nameRaw.toString('ascii');
    return {
        nonce: nonce,
        locktime: locktime,
        name: name,
    };
}
/**
 * Generate a commit transaction.
 * @param coins Array of coins to fund the transaction.
 * @param name Name for name/key pair.
 * @param locktime Absolute locktime
 * @param commitFee Commit fee, in satoshis.
 * @param registerFee Registration fee, in satoshis.
 * @param escrowFee Escrow fee, in satoshis.
 * @param feeRate Fee rate, in satoshis/kilobyte.
 * @param userRing The user's key ring.
 * @param servicePubKey Service public key.
 */
function genCommitTx(coins, name, locktime, commitFee, registerFee, escrowFee, feeRate, userRing, servicePubKey) {
    //
    // Validate name and the service public key
    if (name.length > 64) {
        throw new Error('Name is too long');
    }
    if (!isURISafe(name)) {
        throw new Error('Invalid character(s) in name');
    }
    if (!crypto.secp256k1.publicKeyVerify(servicePubKey)) {
        throw new Error('Invalid service public key');
    }
    // Generate a P2SH address from a redeem script, using a random nonce
    var nonce = randomBytes(32);
    var redeemScript = genCommitRedeemScript(userRing.getPublicKey(), nonce, name, locktime);
    var p2shAddr = genP2shAddr(redeemScript);
    // Generate service address from service public key
    var servicePKH = crypto.hash160(servicePubKey);
    var serviceAddr = address.fromPubkeyhash(servicePKH);
    var lockTx = new mtx();
    // Compute total value of coins
    var total = coins.reduce(function (acc, cur) { return acc + cur.value; }, 0);
    for (var _i = 0, coins_1 = coins; _i < coins_1.length; _i++) {
        var coin = coins_1[_i];
        lockTx.addCoin(coin);
    }
    // Compute change amount
    var changeVal = total - (commitFee + registerFee + escrowFee) - (4 * feeRate);
    // Add nonce OP_RETURN as output 0
    var pubkeyDataScript = script.fromNulldata(nonce);
    lockTx.addOutput(output.fromScript(pubkeyDataScript, 0));
    // Add service upfront fee as output 1
    lockTx.addOutput({
        address: serviceAddr,
        value: commitFee,
    });
    // Add locked fee as output 2
    // Locks up the fee to register, the fee to be put in escrow, and enough for a 4kb tx at current rates
    lockTx.addOutput({
        address: p2shAddr,
        value: registerFee + escrowFee + 4 * feeRate,
    });
    // Add change output as 3
    lockTx.addOutput({
        address: userRing.getAddress(),
        value: changeVal,
    });
    // Add coins as inputs
    for (var i = 0; i < coins.length; ++i) {
        var coin = coins[i];
        lockTx.scriptInput(i, coin, userRing);
    }
    // Each signature is 72 bytes long
    var virtSize = lockTx.getVirtualSize() + coins.length * 72;
    lockTx.subtractIndex(3, Math.ceil(virtSize / 1000 * feeRate));
    // Sign the coins
    for (var i = 0; i < coins.length; ++i) {
        var coin = coins[i];
        lockTx.signInput(i, coin, userRing, script.hashType.ALL);
    }
    return lockTx.toTX();
}
/**
 * Generate a lock transaction.
 * @param commitTX The corresponding commit transaction.
 * @param name The name to use.
 * @param upfrontFee The upfront fee in satoshis to use the service, as
 * determined by the service.
 * @param lockedFee Fee incentivizing registrar to provide service, as
 * determined by the service.
 * @param feeRate Fee rate in satoshi/KB.
 * @param userRing The user key ring.
 * @param servicePubKey Service public key.
 * @param locktime Absolute lock time in blocks.
 */
function genLockTx(commitTX, name, upfrontFee, lockedFee, feeRate, userRing, servicePubKey, locktime) {
    //
    // Input validation
    if (locktime > 500000000) {
        throw new Error('Locktime must be less than 500000000 blocks');
    }
    if (name.length > 64) {
        throw new Error('Name is too long');
    }
    if (!isURISafe(name)) {
        throw new Error('Invalid character(s) in name');
    }
    if (!crypto.secp256k1.publicKeyVerify(servicePubKey)) {
        throw new Error('Invalid service public key');
    }
    if (!verifyCommitTX(commitTX, userRing.getPublicKey(), servicePubKey, name, locktime)) {
        throw new Error('Invalid commitment tx');
    }
    // Generate a P2SH address from redeem script
    var redeemScript = genRedeemScript(userRing.getPublicKey(), servicePubKey, locktime);
    var p2shAddr = genP2shAddr(redeemScript);
    // Generate address from service public key
    var servicePKH = crypto.hash160(servicePubKey);
    var serviceAddr = address.fromPubkeyhash(servicePKH);
    var lockTx = mtx.fromOptions({
        version: 2,
    });
    lockTx.addTX(commitTX, 2);
    lockTx.setSequence(0, 6);
    var total = commitTX.outputs[2].value;
    var changeVal = total - upfrontFee - lockedFee;
    // Add upfront fee as output 0
    lockTx.addOutput({
        address: serviceAddr,
        value: upfrontFee,
    });
    // Add locked fee as output 1
    lockTx.addOutput({
        address: p2shAddr,
        value: lockedFee,
    });
    // Add change output as 2
    lockTx.addOutput({
        address: userRing.getAddress(),
        value: changeVal,
    });
    var nonce = commitTX.outputs[0].script.code[1].data;
    var hashData = serializeCommitData(nonce, locktime, name);
    var commitRedeemScript = genCommitRedeemScript(userRing.getPublicKey(), nonce, name, locktime);
    var unlockScript = new script();
    unlockScript.pushData(hashData);
    unlockScript.pushData(commitRedeemScript.toRaw());
    unlockScript.compile();
    lockTx.inputs[0].script = unlockScript;
    // Add constant for signature
    var virtSize = lockTx.getVirtualSize() + 72;
    // Calculate fee to be paid
    var fee = Math.ceil(virtSize / 1000 * feeRate);
    lockTx.subtractIndex(2, fee);
    // Add signature
    var sig = lockTx.signature(0, commitRedeemScript, total, userRing.getPrivateKey(), script.hashType.ALL, 0);
    unlockScript.insertData(0, sig);
    unlockScript.compile();
    return lockTx.toTX();
}
/**
 * Extract metadata from a script (i.e. the nonce, locktime, and name).
 * @param inputScript Script to read from.
 */
function extractCommitMetadata(inputScript) {
    var meta = deserializeCommitData(inputScript.code[1].data);
    return meta;
}
/**
 * Attempt to extract a name from a lock transaction.
 * @param lockTx Lock transaction to read from.
 * @returns The name if present, and `null` otherwise.
 */
function getLockTxName(lockTx) {
    try {
        var metadata = extractCommitMetadata(lockTx.inputs[0].script);
        return metadata.name;
    }
    catch (e) {
        return null;
    }
}
/**
 * Attempt to extract a time lock from a lock transaction.
 * @param lockTx Lock transaction to read from.
 * @returns The lock time if present, and `null` otherwise.
 */
function getLockTxTime(lockTx) {
    try {
        var metadata = extractCommitMetadata(lockTx.inputs[0].script);
        return metadata.locktime;
    }
    catch (e) {
        return null;
    }
}
/**
 * Attempt to extract a public key from a lock transaction.
 * @param lockTx Lock transaction to read from.
 * @returns The public key if present, and `null` otherwise.
 */
function getLockTxPubKey(lockTx) {
    var inputScript = lockTx.inputs[0].script;
    if (inputScript.code.length < 3) {
        return null;
    }
    var encumberScript = script.fromRaw(inputScript.code[2].data);
    if (encumberScript.code.length < 7) {
        return null;
    }
    var pubKey = encumberScript.code[6].data;
    return pubKey;
}
/**
 * Generate an unlock transaction.
 * @param lockTx The corresponding lock transaction.
 * @param commitTx The corresponding commit transaction.
 * @param feeRate The fee rate in satoshi/KB.
 * @param service Whether to use the script as service.
 * @param ring The service key ring if `service`, otherwise user key ring.
 * @param otherPubKey The user key ring if `service`, otherwise service key ring.
 */
function genUnlockTx(lockTx, commitTx, feeRate, service, ring, otherPubKey) {
    // Disambiguate ring public key and the other public key
    var servicePubKey = service ? ring.getPublicKey() : otherPubKey;
    var userPubKey = !service ? ring.getPublicKey() : otherPubKey;
    //
    // Input validation
    if (!verifyLockTX(lockTx, commitTx, servicePubKey)) {
        throw new BadLockTransactionError();
    }
    if (!crypto.secp256k1.publicKeyVerify(otherPubKey)) {
        throw new Error('Invalid service public key');
    }
    var locktime = getLockTxTime(lockTx);
    if (locktime === null) {
        throw new Error('Could not extract locktime');
    }
    var redeemScript = genRedeemScript(userPubKey, servicePubKey, locktime);
    var val = lockTx.outputs[1].value; // the P2SH output
    var unlockTx = mtx.fromOptions({
        version: 2,
    });
    unlockTx.addTX(lockTx, 1);
    var boolVal = service ? 0 : 1;
    if (service) {
        unlockTx.setLocktime(locktime);
    }
    else {
        unlockTx.setSequence(0, 0);
    }
    unlockTx.addOutput({
        address: ring.getAddress(),
        value: val,
    });
    // Generate new script: <sig> <isUser> <redeemScript>
    var unlockScript = new script();
    unlockScript.pushData(unlockTx.signature(0, redeemScript, val, ring.getPrivateKey(), script.hashType.ALL, 0));
    unlockScript.pushInt(boolVal);
    unlockScript.pushData(redeemScript.toRaw());
    unlockScript.compile();
    unlockTx.inputs[0].script = unlockScript;
    // Compute a fee by multiplying the size by the rate, then account for it
    var virtSize = unlockTx.getVirtualSize();
    var fee = Math.ceil(virtSize / 1000 * feeRate);
    unlockTx.subtractFee(fee);
    // Remake script with the new signature
    var unlockScript2 = new script();
    unlockScript2.pushData(unlockTx.signature(0, redeemScript, val, ring.getPrivateKey(), script.hashType.ALL, 0));
    unlockScript2.pushInt(boolVal);
    unlockScript2.pushData(redeemScript.toRaw());
    unlockScript2.compile();
    unlockTx.inputs[0].script = unlockScript2;
    return unlockTx.toTX();
}

export { genCommitTx, genLockTx, genUnlockTx };
//# sourceMappingURL=bitname.es5.js.map
