"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var bcoin_1 = require("bcoin");
var txs_1 = require("./txs");
/**
 * Returns a Boolean value that indicates if URI has and only has alphanumeric leteral and  '_',  '-', '.' '~'
 * @param str The URI String
 */
function isURISafe(str) {
    var re = /^[a-zA-Z0-9_\-\.\~]*$/;
    return re.test(str);
}
exports.isURISafe = isURISafe;
/**
 * Returns a Boolean value indicates if transaction output is in the correct format of Opcode OP_RETURN.
 * @param output Transaction output
 */
function isValidOP_RETURN(output) {
    // Check that output 0 is an OP_RETURN
    if (!output.script.isNulldata()) {
        return false;
    }
    // Check that output 0 contains 0 satoshis
    if (output.value !== 0) {
        return false;
    }
    // Check that output 0 contains exactly 2 opcodes
    if (output.script.length !== 2) {
        return false;
    }
    return true;
}
/**
 * Returns a Boolean value indicates that whether or not the commit transaction is valid
 * @param tx The commit transaction that needs to be verified.
 * @param userPubKey User's public key
 * @param servicePubKey Service provider's public key
 * @param name A name of at most 64 characters composed of URL-safe characters
 * @param locktime An absolute lock time, in blockheight
 */
function verifyCommitTX(tx, userPubKey, servicePubKey, name, locktime) {
    if (tx.outputs.length < 3) {
        return false;
    }
    // Check that output 0 is an OP_RETURN of the correct form
    if (!isValidOP_RETURN(tx.outputs[0])) {
        return false;
    }
    // Check that output 0 contains a 32-byte nonce
    var nonce = tx.outputs[0].script.code[1].data;
    if (nonce.length !== 32) {
        return false;
    }
    // Check that output 1 is sent to the service's address
    var servicePKH = bcoin_1.crypto.hash160(servicePubKey);
    var serviceAddr = bcoin_1.address.fromPubkeyhash(servicePKH);
    if (tx.outputs[1].getAddress().toBase58('testnet') !== serviceAddr.toBase58('testnet')) {
        return false;
    }
    // Check that output 2 is a P2SH
    if (!tx.outputs[2].script.isScripthash()) {
        return false;
    }
    // Check that output 2 script is correct
    var redeemScript = txs_1.genCommitRedeemScript(userPubKey, nonce, name, locktime);
    var scriptHash = bcoin_1.crypto.hash160(redeemScript.toRaw());
    var p2shAddr = bcoin_1.address.fromScripthash(scriptHash);
    if (tx.outputs[2].getAddress().toBase58('testnet') !== p2shAddr.toBase58('testnet')) {
        return false;
    }
    return true;
}
exports.verifyCommitTX = verifyCommitTX;
/**
 * Returns a Boolean value indicates that whether or not the lock transaction is valid
 * @param tx The lock transaction
 * @param commitTX The corresponding commit transaction.
 * @param servicePubKey service provider's public key
 */
function verifyLockTX(tx, commitTX, servicePubKey) {
    if (tx.outputs.length < 2) {
        return false;
    }
    // Check that input 0 contains a valid pubkey
    var pubKey = txs_1.getLockTxPubKey(tx);
    if (pubKey === null || !bcoin_1.crypto.secp256k1.publicKeyVerify(pubKey)) {
        return false;
    }
    // Check that input 0 name data is only 64 bytes in length
    var nameStr = txs_1.getLockTxName(tx);
    if (nameStr === null) {
        return false;
    }
    if (nameStr.length > 64) {
        return false;
    }
    // Check that output 1 name data contains only URL-safe characters
    if (!isURISafe(nameStr)) {
        return false;
    }
    // Check that output 0 is a P2PKH
    if (!tx.outputs[0].script.isPubkeyhash()) {
        return false;
    }
    // Check that output 0 is sent to the service's address
    var servicePKH = bcoin_1.crypto.hash160(servicePubKey);
    var serviceAddr = bcoin_1.address.fromPubkeyhash(servicePKH);
    if (tx.outputs[0].getAddress().toBase58('testnet') !== serviceAddr.toBase58('testnet')) {
        return false;
    }
    // Check that output 1 is a P2SH
    if (!tx.outputs[1].script.isScripthash()) {
        return false;
    }
    // Check that output 1 script is correct
    var locktime = txs_1.getLockTxTime(tx);
    var redeemScript = txs_1.genRedeemScript(pubKey, servicePubKey, locktime);
    var scriptHash = bcoin_1.crypto.hash160(redeemScript.toRaw());
    var p2shAddr = bcoin_1.address.fromScripthash(scriptHash);
    if (tx.outputs[1].getAddress().toBase58('testnet') !== p2shAddr.toBase58('testnet')) {
        return false;
    }
    // Check that input 0 is a valid commit TX
    if (!verifyCommitTX(commitTX, pubKey, servicePubKey, nameStr, locktime)) {
        return false;
    }
    return true;
}
exports.verifyLockTX = verifyLockTX;
